#include <iostream>
#include <iomanip>
#include <sstream>
#include <string>
#include <stdio.h>
#include <gtkmm/menu.h>
#include <gtkmm/menuitem.h>
#include <gdkmm/cursor.h>
#include <gtkmm/messagedialog.h>
#include <cairomm/context.h>
#include "ChartDrawingArea.h"
#include "Util.h"
#include "METAR.h"
#include "Exception.h"

ChartDrawingArea::ChartDrawingArea(
	BaseObjectType* base_object,
	const Glib::RefPtr<Gnome::Glade::Xml>& glade_xml) : Gtk::DrawingArea(base_object
	)
{
	Cfg=NULL;
	SetOffset(0,0);
	OldWidth=0;
	OldHeight=0;
	Sectional=NULL;
	Statusbar=NULL;
	Dragging=false;
	ContextMenu=NULL;
	RouteData=NULL;
	VolData=NULL;
	ShapefileData=NULL;
	DisplayScale=1.0;
	InterpType=Gdk::INTERP_BILINEAR;
	add_events(Gdk::BUTTON_PRESS_MASK);
	add_events(Gdk::BUTTON_RELEASE_MASK);
	add_events(Gdk::POINTER_MOTION_MASK);
	/* Hint=don't send us another event until we've processed the last one
	 * and called get_pointer()
	 */
	add_events(Gdk::POINTER_MOTION_HINT_MASK);
	LonLatTime=VelTime=SpdTime=AltTime=0;
	ClearLocation();
}

ChartDrawingArea::~ChartDrawingArea() {
	if (Sectional)
		delete Sectional;
}

bool ChartDrawingArea::on_expose_event(GdkEventExpose* ev) {
	GdkRectangle *rects;
	int n_rects,recti;

	if (Sectional && Sectional->GetPixbuf()) {
		/* Recenter if the window size has changed */
		if ((OldWidth && OldHeight) && (OldWidth!=get_width() || OldHeight!=get_height()) ) {
			Shift((int)rint((OldWidth-get_width())/2.0),(int)rint((OldHeight-get_height())/2.0));
			OldWidth=get_width();
			OldHeight=get_height();
		}

		ClipOffset();

		Glib::RefPtr<Gdk::Pixbuf> OutputPixbuf;

		/* If the map has moved 5 pixels up and 5 pixels left it's much better draw those
		 * two 5-pixel wide strips instead of their bounding box (which is the whole window)
		 *
		 * There should be a more C++/gtkmm/gdkmm native way of doing this instead of dragging
		 * out the raw gdk_* functions, but this is the only thing I could get to work.  If this
		 * bothers you, go ahead and fix it.
		 */
		gdk_region_get_rectangles (ev->region, &rects, &n_rects);
		for (recti=0; recti<n_rects;recti++) {
			OutputPixbuf=Gdk::Pixbuf::create(Gdk::COLORSPACE_RGB,true,8,rects[recti].width,rects[recti].height);

			Sectional->GetPixbuf()->scale(
				OutputPixbuf,			/* Output */
				0,						/* dest_x */
				0,						/* dest_y */
				rects[recti].width,					/* dest_width */
				rects[recti].height,				/* dest_height */
				-xoffset-rects[recti].x,						/* offset_x */
				-yoffset-rects[recti].y,						/* offset_y */
				DisplayScale,					/* scale_x */
				DisplayScale,					/* scale_y */
				InterpType
				);

			OutputPixbuf->render_to_drawable(
				get_window(),
				get_style()->get_black_gc(),
				0,	/* src x offset */
				0,	/* src y offset */
				rects[recti].x,			/* dest x offset */
				rects[recti].y,			/* dest y offset */
				OutputPixbuf->get_width(),
				OutputPixbuf->get_height(),
				Gdk::RGB_DITHER_NONE,
				0,
				0);
		}
		g_free(rects);

		/* Only draw the current location if it's been updated in the last 2 seconds */
		if (LonLatTime>time(NULL)-2)
		{
			DrawLocation();
		}
		if (VelTime>time(NULL)-2)
		{
			DrawGroundspeed();
			DrawTrack();
		}
		if (AltTime>time(NULL)-2)
		{
			DrawGPSAltitude();
		}

		if (RouteData)
		{
			DrawGreatCircleRoute(RouteData);
		}

		DrawAirportsWx(VolData);

		DrawShapes(ev);
	}

	return true;
}

void ChartDrawingArea::Shift(int x, int y) {
	SetOffset(xoffset+x,yoffset+y);
}

void ChartDrawingArea::Slide(double x, double y) {
	SetOffset(xoffset+(int)rint(get_width()*x),yoffset+(int)rint(get_height()*y));
}

int ChartDrawingArea::GetXOffset() {
	return xoffset;
}

int ChartDrawingArea::GetYOffset() {
	return yoffset;
}

void ChartDrawingArea::SetOffset(int x, int y) {
	xoffset=x;
	yoffset=y;
	if (Cfg) {
		Cfg->SetChartXOffset(x);
		Cfg->SetChartYOffset(y);
	}
}

void ChartDrawingArea::ClipOffset() {
	if (Sectional) {
		ClipOffset(
			Sectional->GetPixbuf()->get_width(),
			Sectional->GetPixbuf()->get_height(),
			get_width(),
			get_height()
		);
	}
}

void ChartDrawingArea::ClipOffset(int pixwidth, int pixheight, int viewwidth, int viewheight) {
	int newx,newy;

	if (!Sectional) {
		SetOffset(0,0);
		return;
	}

	newx=xoffset;
	newy=yoffset;

	if (xoffset<0)
		newx=0;
	if (yoffset<0)
		newy=0;

	if (newx+viewwidth>(int)rint(pixwidth*DisplayScale))
		newx=(int)rint(pixwidth*DisplayScale)-viewwidth;
	if (newy+viewheight>(int)rint(pixheight*DisplayScale))
		newy=(int)rint(pixheight*DisplayScale)-viewheight;

	SetOffset(newx,newy);
}

/* Project from a raster chart pixel to a screen pixel */
void ChartDrawingArea::ProjectPixel(int *x, int *y) {
	*x=(*x)*DisplayScale-xoffset;
	*y=(*y)*DisplayScale-yoffset;
}

bool ChartDrawingArea::OpenSectional(std::string FileName) {
	SectionalChart *NewSectional;
	NewSectional=new SectionalChart(Cfg,FileName);

	if (!NewSectional->GetValid()) {
		delete NewSectional;

		Gtk::MessageDialog *message = new Gtk::MessageDialog("Unable to open sectional chart file:\n"+FileName,
			false,Gtk::MESSAGE_ERROR,Gtk::BUTTONS_CANCEL,true);
		message->run();
		delete message;

		return false;
	}

	if (Sectional)
		delete Sectional;
	Sectional=NewSectional;
	Sectional->LoadRaster();
	SetOffset(0,0);
	Redraw();
	return true;
}

void ChartDrawingArea::CloseChart() {
	if (Sectional)
		delete Sectional;
	Sectional=NULL;
	SetOffset(0,0);
	Statusbar->pop();
	Redraw();
}

bool ChartDrawingArea::on_motion_notify_event(GdkEventMotion* event) {
	int oldxoffset,oldyoffset;

	if (!Sectional)
		return true;

	if (Dragging) {
		oldxoffset=xoffset;
		oldyoffset=yoffset;
		Shift(DragStartX-event->x,DragStartY-event->y);
		ClipOffset();
		get_window()->scroll(oldxoffset-xoffset,oldyoffset-yoffset);
		DragStartX=event->x;
		DragStartY=event->y;
	} else {
		PutPositionInStatusbar();
	}

	return true;
}

void ChartDrawingArea::SetStatusbar(Gtk::Statusbar *Statusbar) {
	this->Statusbar=Statusbar;
}

void ChartDrawingArea::Redraw() {
	Gdk::Rectangle rect(0,0,get_width(),get_height());
	get_window()->invalidate_rect(rect,1);

	/* Clip the offset so that other classes can read the corrected pointer position */
	ClipOffset();
}

void ChartDrawingArea::PutPositionInStatusbar() {
	if (Statusbar) {
		double Lat,Lon;
		int PointerX,PointerY;
		std::string lonlatstr;

		get_pointer(PointerX,PointerY);
		PointerX+=xoffset;
		PointerY+=yoffset;

		PointerX=(int)rint(PointerX/DisplayScale);
		PointerY=(int)rint(PointerY/DisplayScale);

		Sectional->PixelToLonLat(PointerX,PointerY,&Lon,&Lat);

		LonLatToDMMSSPSSSS(Lon,Lat,&lonlatstr);

		Statusbar->pop();
		Statusbar->push(lonlatstr);
	}
}

void ChartDrawingArea::SetDisplayScale(double scale) {
	double width2,height2;
	double xcenter,ycenter;

	if (scale<0.3)
		scale=0.3;
	if (scale>4.0)
		scale=4.0;

	width2=get_width()/2.0;
	height2=get_height()/2.0;

	xcenter=xoffset+width2;
	ycenter=yoffset+height2;;

	xcenter=(xcenter*scale)/DisplayScale;
	ycenter=(ycenter*scale)/DisplayScale;

	SetOffset((int)rint(xcenter-width2),(int)rint(ycenter-height2));

	DisplayScale=scale;
	if (Cfg) {
		Cfg->SetChartScale(DisplayScale);
	}

	ClipOffset();
}

double ChartDrawingArea::GetDispalyScale() {
	return DisplayScale;
}

void ChartDrawingArea::ScaleDisplayScale(double adjustment) {
	SetDisplayScale(DisplayScale*adjustment);
}

void ChartDrawingArea::SetInterpType(Gdk::InterpType InterpType) {
	this->InterpType=InterpType;
}

void ChartDrawingArea::LonLatToDMMSSPSSSS(double Lon, double Lat, std::string *str) {
	std::ostringstream Text;
	std::string NS,EW;
	int LatDeg,LonDeg,LatMin,LonMin;
	double LatSec,LonSec;

	if (Lat<0)
		NS="S";
	else
		NS="N";
	if (Lon<0)
		EW="W";
	else
		EW="E";
	Lat=fabs(Lat);
	Lon=fabs(Lon);

	LatDeg=(int)(Lat);
	LonDeg=(int)(Lon);

	Lat-=LatDeg;
	Lon-=LonDeg;

	LatMin=(int)(Lat*60.0);
	LonMin=(int)(Lon*60.0);

	Lat-=LatMin/60.0;
	Lon-=LonMin/60.0;

	LatSec=Lat*3600.0;
	LonSec=Lon*3600.0;

	Text.setf(std::ios::fixed);
	Text<<
		std::setprecision(0)<<std::setw(0)<<LatDeg<<"° "<<
		std::setprecision(0)<<std::setw(2)<<std::setfill('0')<<LatMin<<"´ "<<
		std::setprecision(4)<<std::setw(4+3)<<std::setfill('0')<<LatSec<<"˝ "<<
		NS<<", "<<
		std::setprecision(0)<<std::setw(0)<<LonDeg<<"° "<<
		std::setprecision(0)<<std::setw(2)<<std::setfill('0')<<LonMin<<"´ "<<
		std::setprecision(4)<<std::setw(4+3)<<std::setfill('0')<<LonSec<<"˝ "<<
		EW;

	*str=Text.str();
}

void ChartDrawingArea::LonLatToDMMPMMMMMM(double Lon, double Lat, std::string *str) {
	std::ostringstream Text;
	std::string NS,EW;
	int LatDeg,LonDeg;
	double LatMin,LonMin;

	if (Lat<0)
		NS="S";
	else
		NS="N";
	if (Lon<0)
		EW="W";
	else
		EW="E";
	Lat=fabs(Lat);
	Lon=fabs(Lon);

	LatDeg=(int)(Lat);
	LonDeg=(int)(Lon);

	Lat-=LatDeg;
	Lon-=LonDeg;

	LatMin=Lat*60.0;
	LonMin=Lon*60.0;

	Text.setf(std::ios::fixed);
	Text<<
		std::setprecision(0)<<std::setw(0)<<LatDeg<<"° "<<
		std::setprecision(6)<<std::setw(6+3)<<std::setfill('0')<<LatMin<<"´ "<<
		NS<<", "<<
		std::setprecision(0)<<std::setw(0)<<LonDeg<<"° "<<
		std::setprecision(6)<<std::setw(6+3)<<std::setfill('0')<<LonMin<<"´ "<<
		EW;

	*str=Text.str();
}

void ChartDrawingArea::LonLatToDPDDDDDDD(double Lat, double Lon, std::string *str) {
	std::ostringstream Text;
	Text.setf(std::ios::fixed);
	Text<<std::setprecision(7)<<Lon<<","<<Lat;
	*str=Text.str();
}

void ChartDrawingArea::CenterOnLonLat(double Lon, double Lat) {
	int x,y;
	if (Sectional && !Dragging) {
		Sectional->LonLatToPixel(Lon,Lat,&x,&y);
		SetOffset((int)rint(x*DisplayScale-(get_width()/2)),(int)rint(y*DisplayScale-(get_height()/2)));
		ClipOffset();
		Redraw();
	}
}

bool ChartDrawingArea::on_button_press_event(GdkEventButton * event) {
	if (!Sectional)
		return false;
	if (event->type!=GDK_BUTTON_PRESS)
		return false;

	if (event->button==1) {
		Dragging=true;
		DragStartX=(int)event->x;
		DragStartY=(int)event->y;
		Gdk::Cursor *GrabCursor=new Gdk::Cursor(Gdk::HAND2);
		get_window()->set_cursor(*GrabCursor);
		delete GrabCursor;
		return true;
	} else if (event->button==3) {
		ContextMenu = new Gtk::Menu();
		Gtk::MenuItem *Center=new Gtk::MenuItem("Center");
		ContextMenu->append(*Center);
		ContextMenu->popup(event->button,event->time);

		Center->signal_activate().connect(sigc::bind<int,int>(sigc::mem_fun(*this,&ChartDrawingArea::on_Center_activate),(int)event->x,(int)event->y));

		ContextMenu->show_all();
		return true;
	}
	return false;
}

bool ChartDrawingArea::on_button_release_event(GdkEventButton * event) {
	if (!Sectional)
		return false;
	if (event->type!=GDK_BUTTON_RELEASE)
		return false;

	if (event->button==1) {
		Dragging=false;
		get_window()->set_cursor(); /* Go from the hand back to normal */
		return true;
	} else if (event->button==3) {
		return true;
	}
	return false;
}

void ChartDrawingArea::on_Center_activate(int x, int y) {
	/* Center the chart */
	Shift(x-get_width()/2,y-get_height()/2);
	ClipOffset();

	/* Center the pointer */
	int XLocal,YLocal;
	int XScreen,YScreen;
	Gdk::ModifierType ModifierType;
	get_pointer(XLocal,YLocal);
	get_display()->get_pointer(XScreen,YScreen,ModifierType);
	get_display()->warp_pointer(
		get_screen(),
		get_width()/2+(XScreen-XLocal),
		get_height()/2+(YScreen-YLocal)
		);

	Redraw();
	PutPositionInStatusbar();
}

void ChartDrawingArea::DrawLocation() {
	Cairo::RefPtr<Cairo::Context> Context = get_window()->create_cairo_context();
	int x, y, NorthPointX, NorthPointY;
	double ScreenTrack;

	if (!Sectional)
		return;

	time_t tooold;
	tooold=time(NULL)-2;
	if (LonLatTime<tooold)
		return;

	Sectional->LonLatToPixel(Lon,Lat,&x,&y);

	Context->set_source_rgba(0.2, 0.2, 0.8, 0.4);
	Context->set_line_width(1.0);
	Context->translate(x*DisplayScale-xoffset,y*DisplayScale-yoffset);

	/* Draw the main dot */
	Context->save();
	Context->arc(0.0,0.0,15.0,0.0,2*M_PI);
	Context->fill_preserve();
	Context->stroke();
	Context->restore();

	if (VelTime>tooold) {
		/* Draw a track marker */

		/* North isn't vertical most of the time - pick a point a bit
		 * North, translate it to pixels, then figure out the screen
		 * direction to North using it.
		 */
		Sectional->LonLatToPixel(Lon,Lat+(5.0/60.0),&NorthPointX,&NorthPointY);
		/* Figure the screen track to true North */
		ScreenTrack=atan(((double)NorthPointY-y)/((double)NorthPointX-x));
		/* Account for the vagaries of atan() */
		if (ScreenTrack<0.0)
			ScreenTrack+=M_PI/2.0;
		else
			ScreenTrack-=M_PI/2.0;

		/* Add the desired track for the final result */
		ScreenTrack+=Trk/(180.0/M_PI);	/* Convert to radians */

		/* Draw a Track nub */
		Context->save();
		Context->rotate(ScreenTrack);
		Context->arc(0.0,-15.0,3.0,0.0,2*M_PI);
		Context->fill_preserve();
		Context->stroke();
		Context->restore();

		/* Draw a Track wedge */
		Context->save();
		Context->set_line_width(5.0);
		Context->rotate(ScreenTrack);
		Context->move_to(0.0,-18.0);
		Context->line_to(-10.0,-200.0);
		Context->line_to(10.0,-200.0);
		Context->line_to(0.0,-18.0);
		Context->stroke();
		Context->restore();
	}
}

void ChartDrawingArea::DrawGroundspeed() {
	Cairo::RefPtr<Cairo::Context> Context = get_window()->create_cairo_context();
	std::ostringstream ValueOStr;
	std::string LabelStr;
	Cairo::TextExtents ValueExtents,LabelExtents;
	double TextWidth,TextHeight,XBearing,YBearing;

	if (!Sectional)
		return;

	time_t tooold;
	tooold=time(NULL)-2;
	if (VelTime<tooold)
		return;

	/* Prepare and size the text */
	ValueOStr<<(int)round(MPERS2KT(Spd));
	LabelStr="kt";

	Context->set_font_size(DataFieldValueSize);
	Context->get_text_extents(ValueOStr.str(),ValueExtents);
	Context->set_font_size(DataFieldLabelSize);
	Context->get_text_extents(LabelStr,LabelExtents);

	TextWidth=ValueExtents.width+LabelExtents.width;
	if (LabelExtents.height>ValueExtents.height)
		TextHeight=LabelExtents.height;
	else
		TextHeight=ValueExtents.height;

	XBearing=ValueExtents.x_bearing+ValueExtents.x_bearing;
	if (LabelExtents.y_bearing<ValueExtents.y_bearing)
		YBearing=LabelExtents.y_bearing;
	else
		YBearing=ValueExtents.y_bearing;

	/* Draw the background */
	Context->save();
	Context->set_line_join(Cairo::LINE_JOIN_ROUND);
	Context->set_line_width(TextHeight);
	Context->set_source_rgba(1.0, 1.0, 1.0, 0.7);
	Context->rectangle(0,0,TextWidth+DataFieldPadding,TextHeight);
	Context->stroke();
	Context->restore();

	/* Draw the text/foreground */
	Context->save();
	Context->set_source_rgba(0.0, 0.0, 0.0, 0.3);

	Context->translate(
			DataFieldPadding-XBearing,
			TextHeight);

	Context->set_font_size(DataFieldValueSize);
	Context->show_text(ValueOStr.str());

	Context->set_font_size(DataFieldLabelSize);
	Context->show_text(LabelStr);

	Context->stroke();
	Context->restore();
}

void ChartDrawingArea::DrawGPSAltitude() {
	Cairo::RefPtr<Cairo::Context> Context = get_window()->create_cairo_context();
	std::ostringstream ValueOStr;
	std::string LabelStr;
	Cairo::TextExtents ValueExtents,LabelExtents;
	double TextWidth,TextHeight,XBearing,YBearing;

	if (!Sectional)
		return;

	time_t tooold;
	tooold=time(NULL)-2;
	if (VelTime<tooold)
		return;

	/* Prepare and size the text */
	ValueOStr<<(int)round(METERS2FT(Alt));
	LabelStr="ft";

	Context->set_font_size(DataFieldValueSize);
	Context->get_text_extents(ValueOStr.str(),ValueExtents);
	Context->set_font_size(DataFieldLabelSize);
	Context->get_text_extents(LabelStr,LabelExtents);

	TextWidth=ValueExtents.width+LabelExtents.width;
	if (LabelExtents.height>ValueExtents.height)
		TextHeight=LabelExtents.height;
	else
		TextHeight=ValueExtents.height;

	XBearing=ValueExtents.x_bearing+ValueExtents.x_bearing;
	if (LabelExtents.y_bearing<ValueExtents.y_bearing)
		YBearing=LabelExtents.y_bearing;
	else
		YBearing=ValueExtents.y_bearing;

	/* Draw the background */
	Context->save();
	Context->set_line_join(Cairo::LINE_JOIN_ROUND);
	Context->set_line_width(TextHeight);
	Context->set_source_rgba(1.0, 1.0, 1.0, 0.7);
	Context->rectangle(get_width()-(TextWidth+DataFieldPadding),0,TextWidth,TextHeight);
	Context->stroke();
	Context->restore();

	/* Draw the text/foreground */
	Context->save();
	Context->set_source_rgba(0.0, 0.0, 0.0, 0.3);

	Context->translate(
			get_width()-(DataFieldPadding+XBearing+TextWidth),
			TextHeight);

	Context->set_font_size(DataFieldValueSize);
	Context->show_text(ValueOStr.str());

	Context->set_font_size(DataFieldLabelSize);
	Context->show_text(LabelStr);

	Context->stroke();
	Context->restore();
}

void ChartDrawingArea::DrawTrack() {
	Cairo::RefPtr<Cairo::Context> Context = get_window()->create_cairo_context();
	std::ostringstream ValueOStr;
	std::string LabelStr;
	Cairo::TextExtents ValueExtents,LabelExtents;
	double TextWidth,TextHeight,XBearing,YBearing;

	if (!Sectional)
		return;

	time_t tooold;
	tooold=time(NULL)-2;
	if (VelTime<tooold)
		return;

	/* Prepare and size the text */
	ValueOStr<<(int)round(Trk);
	LabelStr="DegT";

	Context->set_font_size(DataFieldValueSize);
	Context->get_text_extents(ValueOStr.str(),ValueExtents);
	Context->set_font_size(DataFieldLabelSize);
	Context->get_text_extents(LabelStr,LabelExtents);

	TextWidth=ValueExtents.width+LabelExtents.width;
	if (LabelExtents.height>ValueExtents.height)
		TextHeight=LabelExtents.height;
	else
		TextHeight=ValueExtents.height;

	XBearing=ValueExtents.x_bearing+ValueExtents.x_bearing;
	if (LabelExtents.y_bearing<ValueExtents.y_bearing)
		YBearing=LabelExtents.y_bearing;
	else
		YBearing=ValueExtents.y_bearing;

	/* Draw the background */
	Context->save();
	Context->set_line_join(Cairo::LINE_JOIN_ROUND);
	Context->set_line_width(TextHeight);
	Context->set_source_rgba(1.0, 1.0, 1.0, 0.7);
	Context->rectangle(get_width()/2-(TextWidth)/2,0,TextWidth,TextHeight);
	Context->stroke();
	Context->restore();

	/* Draw the text/foreground */
	Context->save();
	Context->set_source_rgba(0.0, 0.0, 0.0, 0.3);

	Context->translate(
			get_width()/2-(XBearing+TextWidth)/2,
			TextHeight);

	Context->set_font_size(DataFieldValueSize);
	Context->show_text(ValueOStr.str());

	Context->set_font_size(DataFieldLabelSize);
	Context->show_text(LabelStr);

	Context->stroke();
	Context->restore();
}

void ChartDrawingArea::MarkLocation(double Lon, double Lat, double *Trk, double *Spd, double *Alt) {
	LonLatTime=time(NULL);
	this->Lon=Lon;
	this->Lat=Lat;

	if (Trk && Spd) {
		VelTime=LonLatTime;
		this->Trk=*Trk;
		this->Spd=*Spd;
	}
	if (Alt) {
		AltTime=LonLatTime;
		this->Alt=*Alt;
	}

	Redraw();
}

void ChartDrawingArea::ClearLocation() {
	LonLatTime=0;
	VelTime=0;
}

void ChartDrawingArea::SetRoute(Route *RouteData)
{
		this->RouteData=RouteData;
}

void ChartDrawingArea::SetVolData(Volume *VolData)
{
	this->VolData=VolData;
}

void ChartDrawingArea::SetShapefile(Shapefile *ShapefileData)
{
	this->ShapefileData=ShapefileData;
}

std::string ChartDrawingArea::GetCurrSectionalFileName()
{
	if (!Sectional)
		return "";
	return Sectional->GetFileName();
}

SectionalChart *ChartDrawingArea::GetCurrSectionalChart() {
	return Sectional;
}

Glib::RefPtr<Gdk::Window> ChartDrawingArea::GetWindow() {
	return get_window();
}

void ChartDrawingArea::DrawRhumbLineRoute(Route *RouteData)
{
	int XPixel,YPixel;
	std::list<RoutePoint*> *RoutePoints,*LinRoutePoints;
	std::list<RoutePoint*>::iterator RoutePointi;

	if (!Sectional) {
		/* Nothing to draw on */
		return;
	}

	RoutePoints=RouteData->GetRoutePoints();
	if (RoutePoints->size()<2) {
		/* Nothing to draw */
		return;
	}

	Cairo::RefPtr<Cairo::Context> Context = get_window()->create_cairo_context();

	/* Draw direct (non-great-circle) route */

	/* Set up brush */
	Context->set_source_rgba(0.5, 0.2, 0.8, 0.4);
	Context->set_line_width(5.0);

	/* Go to each point */
	for (RoutePointi=RoutePoints->begin();RoutePointi!=RoutePoints->end();RoutePointi++)
	{
		Sectional->LonLatToPixel((*RoutePointi)->GetLon(),(*RoutePointi)->GetLat(),&XPixel,&YPixel);

		/* Convert from raw pixels to pixels as displayed (scaled+shifted) */
		ProjectPixel(&XPixel,&YPixel);

		if (RoutePointi==RoutePoints->begin()) {
			/* Go to the start position */
			Context->move_to(XPixel,YPixel);
		} else {
			/* Go to this position */
			Context->line_to(XPixel,YPixel);
		}
	}

	/* Draw the line */
	Context->stroke();

	/* Draw great circle route */
	LinRoutePoints=RouteData->GetLinearRoutePoints();

	/* Set up brush */
	Context->set_source_rgba(0.5, 0.2, 0.8, 0.4);
	Context->set_line_width(5.0);

	/* Go to each point */
	for (RoutePointi=LinRoutePoints->begin();RoutePointi!=LinRoutePoints->end();RoutePointi++)
	{
		Sectional->LonLatToPixel((*RoutePointi)->GetLon(),(*RoutePointi)->GetLat(),&XPixel,&YPixel);

		/* Convert from raw pixels to pixels as displayed (scaled+shifted) */
		ProjectPixel(&XPixel,&YPixel);

		if (RoutePointi==RoutePoints->begin()) {
			/* Go to the start position */
			Context->move_to(XPixel,YPixel);
		} else {
			/* Go to this position */
			Context->line_to(XPixel,YPixel);
		}
	}

	/* Draw the line */
	Context->stroke();
}

void ChartDrawingArea::DrawGreatCircleRoute(Route *RouteData)
{
	int XPixel,YPixel;
	std::list<RoutePoint*> *LinRoutePoints;
	std::list<RoutePoint*>::iterator RoutePointi;

	if (!RouteData) {
		/* Nothing to draw */
		return;
	}

	if (!Sectional) {
		/* Nothing to draw on */
		return;
	}

	LinRoutePoints=RouteData->GetLinearRoutePoints();
	if (LinRoutePoints->size()<2) {
		/* Nothing to draw */
		return;
	}

	Cairo::RefPtr<Cairo::Context> Context = get_window()->create_cairo_context();

	/* Set up brush */
	Context->set_source_rgba(0.8, 0.2, 0.5, 0.4);
	Context->set_line_width(5.0);

	/* Go to each point */
	for (RoutePointi=LinRoutePoints->begin();RoutePointi!=LinRoutePoints->end();RoutePointi++)
	{
		Sectional->LonLatToPixel((*RoutePointi)->GetLon(),(*RoutePointi)->GetLat(),&XPixel,&YPixel);

		/* Convert from raw pixels to pixels as displayed (scaled+shifted) */
		ProjectPixel(&XPixel,&YPixel);

		if (RoutePointi==LinRoutePoints->begin()) {
			/* Go to the start position */
			Context->move_to(XPixel,YPixel);
		} else {
			/* Go to this position */
			Context->line_to(XPixel,YPixel);
		}
	}

	/* Draw the line */
	Context->stroke();
}

void ChartDrawingArea::DrawAirportsWx(Volume *Voldata) {
	double East,West,North,South;

	if (!VolData & !Sectional)
		return;

	Cairo::RefPtr<Cairo::Context> Context = get_window()->create_cairo_context();

	/* Set up brush */
	Context->set_source_rgba(0.2, 0.8, 0.2, 0.8);

	Sectional->GetBounds(&East,&West,&North,&South);

	for (std::vector<Airport*>::iterator Airporti=VolData->GetAirports()->begin();Airporti!=VolData->GetAirports()->end();++Airporti) {
		if (
			(*Airporti)->GetMETARs()->size() &&
			(*Airporti)->GetLon()<East &&
			(*Airporti)->GetLon()>West &&
			(*Airporti)->GetLat()<North &&
			(*Airporti)->GetLat()>South
			) {
			DrawAirportWx(Context, *Airporti);
		}
	}
}

/*
 * Useful links include:
 * http://adds.aviationweather.gov/metars/description_ifr.php
 * http://adds.aviationweather.gov/metars/description2.php
 */
void ChartDrawingArea::DrawAirportWx(Cairo::RefPtr<Cairo::Context> Context, Airport *AirportData) {
	int XPixel,YPixel;
	METAR *LatestWx=AirportData->GetMETARs()->back();
	METAR::CeilingCover_enum CeilingCover;

	if (!Sectional)
		return;
	if (!LatestWx)
		return;
	if (LatestWx->GetCeilingCover()==METAR::CEILINGCOVER_INVALID)
		return;

	Context->save();
	Sectional->LonLatToPixel(AirportData->GetLon(),AirportData->GetLat(),&XPixel,&YPixel);

	/* Convert from raw pixels to pixels as displayed (scaled+shifted) */
	ProjectPixel(&XPixel,&YPixel);

	Context->set_source_rgba(1.0, 1.0, 1.0, 0.4);
	Context->arc(XPixel,YPixel,6.0,0.0,2*M_PI);
	Context->fill();

	switch (LatestWx->GetFR()) {
		case METAR::FR_VFR:
			Context->set_source_rgba(0.0, 1.0, 0.0, 1.0);	/* Green */
			break;
		case METAR::FR_MVFR:
			Context->set_source_rgba(0.0, 0.0, 1.0, 1.0);	/* Blue */
			break;
		case METAR::FR_IFR:
			Context->set_source_rgba(1.0, 0.0, 0.0, 1.0);	/* Red */
			break;
		case METAR::FR_LIFR:
			Context->set_source_rgba(1.0, 0.0, 1.0, 1.0);	/* Magenta */
			break;
		default:
			Context->set_source_rgba(0.0, 0.0, 0.0, 0.0);	/* Should never happen */
	}

	CeilingCover=LatestWx->GetCeilingCover();

	if (CeilingCover==METAR::CEILINGCOVER_MISSING) {
		/* Draw a crude M for Missing */
		Context->set_source_rgba(0.0, 0.0, 0.0, 1.0);	/* Black */
		Context->set_line_width(2.0);
		Context->move_to(XPixel-5,YPixel+5);
		Context->line_to(XPixel-5,YPixel-5);
		Context->line_to(XPixel,YPixel+3);
		Context->line_to(XPixel+5,YPixel-5);
		Context->line_to(XPixel+5,YPixel+5);
		Context->stroke();
	} else if (CeilingCover==METAR::CEILINGCOVER_CLEAR) {
		Context->set_line_width(2.0);
		Context->rectangle(XPixel-6,YPixel-6,12,12);
		Context->stroke();
	} else {
		Context->set_line_width(2.0);
		Context->arc(XPixel,YPixel,6.0,0.0,2*M_PI);
		Context->stroke();
		Context->set_line_width(1.0);

		switch (CeilingCover) {
			case METAR::CEILINGCOVER_FEW:
				Context->move_to(XPixel,YPixel-5);
				Context->line_to(XPixel,YPixel+5);
				Context->stroke();
				break;
			case METAR::CEILINGCOVER_SCATTERED:
				Context->move_to(XPixel,YPixel);
				Context->arc(XPixel,YPixel,6.0,2.0*M_PI*0.75,2.0*M_PI);
				Context->line_to(XPixel,YPixel);
				Context->close_path();
				Context->fill();
				break;
			case METAR::CEILINGCOVER_BROKEN:
				Context->move_to(XPixel,YPixel);
				Context->arc(XPixel,YPixel,6.0,2.0*M_PI*(-0.25),2.0*M_PI*0.5);
				Context->line_to(XPixel,YPixel);
				Context->close_path();
				Context->fill();
				break;
			case METAR::CEILINGCOVER_OVERCAST:
				Context->arc(XPixel,YPixel,6.0,0.0,2.0*M_PI);
				Context->fill();
				break;
			case METAR::CEILINGCOVER_OBSCURED:
				Context->move_to(XPixel-5,YPixel-5);
				Context->line_to(XPixel+5,YPixel+5);
				Context->move_to(XPixel+5,YPixel-5);
				Context->line_to(XPixel-5,YPixel+5);
				Context->stroke();
				break;
			default:
				/* Ignore - it's been handled elsewhere */
				break;
		}
	}
	Context->restore();
}

void ChartDrawingArea::SetConfig(Config *Cfg) {
	this->Cfg=Cfg;
}

void ChartDrawingArea::DrawShapes(GdkEventExpose* ev) {
	int XPixel, YPixel;
	double NB,SB,EB,WB;
	GdkRectangle *Rects;
	int nRects, Recti;

	if (!ShapefileData)
		return;

	gdk_region_get_rectangles(ev->region, &Rects, &nRects);
	Cairo::RefPtr<Cairo::Context> Context = get_window()->create_cairo_context();

	/* Set up brush */
	Context->set_source_rgba(1.0, 0.0, 0.0, 1.0);
	Context->set_line_width(1.0);

	for (Recti=0;Recti<nRects;Recti++) {
		/* Expand our clipping rect a little bit to account for nonlinearities in
		 * our Lat/Lon -> X/Y mapping.
		 */
		Rects[Recti].x-=50;
		Rects[Recti].y-=50;
		Rects[Recti].width+=100;
		Rects[Recti].height+=100;
		Sectional->PixelToLonLat((Rects[Recti].x+xoffset)/DisplayScale, (Rects[Recti].y+yoffset)/DisplayScale, &WB, &NB);
		Sectional->PixelToLonLat((Rects[Recti].x+Rects[Recti].width+xoffset)/DisplayScale, (Rects[Recti].y+Rects[Recti].height+yoffset)/DisplayScale, &EB, &SB);
		for (std::vector<ShapefilePolyLine*>::iterator PolyLinei=ShapefileData->GetPolyLines()->begin(); PolyLinei!=ShapefileData->GetPolyLines()->end(); PolyLinei++) {
			if ((*PolyLinei)->ContainedIn(SB,WB,NB,EB)) {
				for (std::vector<ShapefilePolyLinePart*>::iterator Parti=(*PolyLinei)->GetParts()->begin(); Parti!=(*PolyLinei)->GetParts()->end(); Parti++) {
					int Contig=0;
					for (std::list<struct shapefile_point_struct*>::iterator Pointi=(*Parti)->GetPoints()->begin(); Pointi!=(*Parti)->GetPoints()->end(); Pointi++) {
						if ((*Pointi)->x>=WB && (*Pointi)->x<=EB && (*Pointi)->y>=SB && (*Pointi)->y<=NB) {
							/* Parts of Alaska go from -180 lon to +180 lon, handle them as a special case */
							if ((*Pointi)->x>0)
								(*Pointi)->x*=-1;

							Sectional->LonLatToPixel((*Pointi)->x,(*Pointi)->y,&XPixel,&YPixel);

							/* Convert from raw pixels to pixels as displayed (scaled+shifted) */
							ProjectPixel(&XPixel,&YPixel);

							if (XPixel>=Rects[Recti].x && XPixel<=(Rects[Recti].x+Rects[Recti].width) && YPixel>=Rects[Recti].y && YPixel<=(Rects[Recti].y+Rects[Recti].height)) {
								/* We're back into the visible area - start drawing from the previous point so we go all the way to the edge */
								if (Contig==0 && Pointi!=(*Parti)->GetPoints()->begin()) {
									Pointi--;	/* Go back to the previous point */
									int LastXPixel, LastYPixel;
									if ((*Pointi)->x>0)
										(*Pointi)->x*=-1;

									Sectional->LonLatToPixel((*Pointi)->x,(*Pointi)->y,&LastXPixel,&LastYPixel);

									/* Convert from raw pixels to pixels as displayed (scaled+shifted) */
									ProjectPixel(&LastXPixel,&LastYPixel);
									Context->move_to(LastXPixel, LastYPixel);
									Pointi++;
								}
								Context->line_to(XPixel,YPixel);
								Contig=1;
							} else {
								Contig=0;
							}
						} else {
							Contig=0;
						}
					}
				}
				Context->stroke();
			}
		}
	}
	g_free(Rects);
}
