#include <fstream>
#include <vector>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include "ARINC424_18.h"
#include "NFD.h"

/* The NFD is published using ARINC 424 rev 18.
 */

NFD::NFD()
{
}

bool NFD::ParseHeader(std::ifstream *IFStream) {
	std::string Line;
	std::string Magic1,Magic2,Format;

	getline(*IFStream,Line);
	Magic1=Line.substr(0,3);
	if (Magic1.compare("VOL")!=0)
		return false;	/* Bad magic */
	Magic2=Line.substr(37,3);
	if (Magic2.compare("FAA")!=0)
		return false;	/* Bad magic */
	Volume=Line.substr(6,4);
	getline(*IFStream,Line);
	Format=Line.substr(4,8);
	if (Format.compare("FAANFD18")!=0)
		return false;	/* Improper data format */

	/* Ignore the rest */
	getline(*IFStream,Line);
	getline(*IFStream,Line);

	return true;
}

void NFD::StoreLine(std::string Line) {
	/* Load the lines we care about into memory.  We don't parse them directly because we
	 * want to process the information in a different order than it appears.
	 */
	if (strcmp(arinc424_18_parse_recordtype(Line.c_str()),ARINC424_18_RECTYPE_AIRPORT)==0)
		AirportLines.push_back(Line);
	else if (strcmp(arinc424_18_parse_recordtype(Line.c_str()),ARINC424_18_RECTYPE_AIRPORT_RUNWAY)==0)
		RunwayLines.push_back(Line);
	else if (strcmp(arinc424_18_parse_recordtype(Line.c_str()),ARINC424_18_RECTYPE_NAVAID_VHF)==0)
		VORLines.push_back(Line);
	else if (strcmp(arinc424_18_parse_recordtype(Line.c_str()),ARINC424_18_RECTYPE_NAVAID_NDB)==0)
		NDBLines.push_back(Line);
	else if (strcmp(arinc424_18_parse_recordtype(Line.c_str()),ARINC424_18_RECTYPE_ENROUTE_WAYPOINT)==0)
		EnrouteWaypointLines.push_back(Line);
}

void NFD::FreeLines() {
	AirportLines.clear();
	RunwayLines.clear();
	VORLines.clear();
	NDBLines.clear();
	EnrouteWaypointLines.clear();
}

bool NFD::ParseAirportLine(std::vector<std::string>::iterator AirportLinei, std::vector<Airport*> *Airports) {
	Airport *AP = new Airport;
	if (AP->ParseNFD(*AirportLinei)) {
		Airports->push_back(AP);
		return true;
	} else {
		delete AP;
		return false;
	}
	return false;
}

bool NFD::ParseRunwayLine(std::vector<std::string>::iterator RunwayLinei, std::vector<Airport*> *Airports) {
	std::vector<Runway*> Runways;
	Airport *A=NULL;
	static Airport *LastA=NULL;
	Runway *R = new Runway;
	std::vector<Airport*>::iterator Airporti;

	if (R->ParseNFD(*RunwayLinei)) {
		A=NULL;
		if (LastA)
		{
			/* If the last airport was the one with this runway, use it */
			if (LastA->GetICAO().compare(R->GetAirportICAO())==0)
				A=LastA;
		}
		
		/* If necessary, do a complete search for our airport */
		if (!A) {
			for(Airporti = Airports->begin(); Airporti != Airports->end(); Airporti++) {
				if ((*Airporti)->GetICAO()==R->GetAirportICAO())
				{
					A=*Airporti;
					break;
				}
			}
			if (Airporti==Airports->end()) {	/* No match found */
				return false;
			}
		}

		/* We found our airport, link with it */
		if (A)
		{
			A->AddRunway(R);
			/* The current airport is likely to be the next airport we need - cache it */
			LastA=A;
		}
		return true;
	} else {
		return false;
	}

	return false;
}

bool NFD::ParseVORLine(std::vector<std::string>::iterator VORLinei, std::vector<NavAid*> *NavAids) {
	VOR *V = new VOR;
	if (V->ParseNFD(*VORLinei)) {
		NavAids->push_back(V);
		return true;
	} else {
		delete V;
		return false;
	}
	return false;
}

bool NFD::ParseNDBLine(std::vector<std::string>::iterator NDBLinei, std::vector<NavAid*> *NavAids) {
	NDB *N = new NDB;
	if (N->ParseNFD(*NDBLinei)) {
		NavAids->push_back(N);
		return true;
	} else {
		delete N;
		return false;
	}
	return false;
}

bool NFD::ParseEnrouteWaypointLine(std::vector<std::string>::iterator EnrouteWaypointLinei, std::vector<EnrouteWaypoint*> *EnrouteWaypoints) {
	EnrouteWaypoint *E = new EnrouteWaypoint;
	if (E->ParseNFD(*EnrouteWaypointLinei)) {
		EnrouteWaypoints->push_back(E);
		return true;
	} else {
		delete E;
		return false;
	}
	return false;
}



NFD::~NFD()
{
}

std::vector<std::string> *NFD::GetAirportLines() {
	return &AirportLines;
}

std::vector<std::string> *NFD::GetRunwayLines() {
	return &RunwayLines;
}

std::vector<std::string> *NFD::GetVORLines() {
	return &VORLines;
}

std::vector<std::string> *NFD::GetNDBLines() {
	return &NDBLines;
}

std::vector<std::string> *NFD::GetEnrouteWaypointLines() {
	return &EnrouteWaypointLines;
}

std::string NFD::GetVolume() {
	return Volume;
}
