#ifndef SHAPEFILE_H_
#define SHAPEFILE_H_

/* <cstdint> is currently experimental (according to Fedora 10), so use the C one */
#include <stdint.h>
#include <cstdio>
#include <vector>
#include <list>
#include <string>

#define SHAPEFILE_SWAP32(x) ( (x&0xFF000000)>>24 | (x&0x00FF0000)>>8 | (x&0x0000FF00)<<8 | (x&0x000000FF)<<24 )
#define SHAPEFILE_NODATA 1e-38
/* Version and File Code from the ESRI Shapefile Technical Description dated July 1998 */
#define SHAPEFILE_VERSION 1000
#define SHAPEFILE_FILECODE 9994

/* All integers are 32-bit signed bits.  _b suffix=Big Endian, _l suffix=Little Endian. */
/* All floats are 64-bit IEEE double-precisions floats, little endian. */

enum shapefile_shapetype_enum {
	SHAPETYPE_NULL,
	SHAPETYPE_POINT,
	SHAPETYPE_POLYLINE=3,
	SHAPETYPE_POLYGON=5,
	SHAPETYPE_MULTIPOINT=8,
	SHAPETYPE_POINTZ=11,
	SHAPETYPE_POLYLINEZ=13,
	SHAPETYPE_POLYGONZ=15,
	SHAPETYPE_MULTIPOINTZ=18,
	SHAPETYPE_POINTM=21,
	SHAPETYPE_POLYLINEM=23,
	SHAPETYPE_POLYGONM=25,
	SHAPETYPE_MULTIPOINTM=28,
	SHAPETYPE_MULTIPATCH=31
};

/* First 100 bytes of *.shp file */
struct shapefile_fileheader_struct {
	int32_t filecode_b;
	int32_t unused_b[5];
	int32_t filelength_b;	/* In 16-byte words */
	int32_t version_l;
	int32_t shapetype_l;
	double xmin;
	double ymin;
	double xmax;
	double ymax;
	double zmin;
	double zmax;
	double mmin;
	double mmax;
} __attribute__ ((packed));

/* First 100 bytes of *.shx file */
struct shapefile_indexfileheader_struct {
	int32_t filecode_b;
	int32_t unused_b[5];
	int32_t filelength_b;	/* In 16-byte words */
	int32_t version_l;
	int32_t shapetype_l;
	double xmin;
	double ymin;
	double xmax;
	double ymax;
	double zmin;
	double zmax;
	double mmin;
	double mmax;
} __attribute__ ((packed));

struct shapefile_point_struct {
	double x;
	double y;
} __attribute__ ((packed));

struct shapefile_pointm_struct {
	double x;
	double y;
	double m;
} __attribute__ ((packed));

struct shapefile_pointz_struct {
	double x;
	double y;
	double z;
	double m;
} __attribute__ ((packed));

struct shapefile_recordheader_struct {
	int32_t recordnumber_b;
	int32_t contentlength_b;
} __attribute__ ((packed));

struct shapefile_recordindex_struct {
	int32_t offset_b;
	int32_t contentlength_b;
} __attribute__ ((packed));

struct shapefile_record_struct {
	int32_t shapetype_l;
	union {
		struct {} null;

		struct shapefile_point_struct point;
		struct {
			double box[4];
			int32_t numpoints;
			struct shapefile_point_struct *points;		/* array of size numpoints */
		} multipoint __attribute__ ((packed));
		struct {
			double box[4];
			int32_t numparts;
			int32_t numpoints;
			int32_t *parts;					/* array of size numparts */
			struct shapefile_point_struct *points;		/* array of size numpoints */
		} polyline __attribute__ ((packed));
		struct {
			double box[4];
			int32_t numparts;
			int32_t numpoints;
			int32_t *parts;					/* array of size numparts */
			struct shapefile_point_struct *points;		/* array of size numpoints */
		} polygon __attribute__ ((packed));

		struct shapefile_pointm_struct pointm;
		struct {
			double box[4];
			int32_t numpoints;
			struct shapefile_pointm_struct *points;		/* array of size numpoints */
			double mmin;
			double mmax;
			double *ms;					/* array of size numpoints */
		} multipointm __attribute__ ((packed));
		struct {
			double box[4];
			int32_t numparts;
			int32_t numpoints;
			int32_t *parts;					/* array of size numparts */
			struct shapefile_point_struct *points;		/* array of size numpoints */
			double mmin;
			double mmax;
			double *ms;					/* array of size numpoints */
		} polylinem __attribute__ ((packed));
		struct {
			double box[4];
			int32_t numparts;
			int32_t numpoints;
			int32_t *parts;					/* array of size numparts */
			struct shapefile_point_struct *points;		/* array of size numpoints */
			double mmin;
			double mmax;
			double *ms;					/* array of size numpoints */
		} polygonm __attribute__ ((packed));

		struct shapefile_pointz_struct pointz;
		struct {
			double box[4];
			int32_t numpoints;
			struct shapefile_pointm_struct *points;		/* array of size numpoints */
			double zmin;
			double zmax;
			double *zs;					/* array of size numpoints */
			double mmin;
			double mmax;
			double *ms;					/* array of size numpoints */
		} multipointz __attribute__ ((packed));
		struct {
			double box[4];
			int32_t numparts;
			int32_t numpoints;
			int32_t *parts;					/* array of size numparts */
			struct shapefile_point_struct *points;		/* array of size numpoints */
			double zmin;
			double zmax;
			double *zs;					/* array of size numpoints */
			double mmin;
			double mmax;
			double *ms;					/* array of size numpoints */
		} polylinez __attribute__ ((packed));
		struct {
			double box[4];
			int32_t numparts;
			int32_t numpoints;
			int32_t *parts;					/* array of size numparts */
			struct shapefile_point_struct *points;		/* array of size numpoints */
			double zmin;
			double zmax;
			double *zs;					/* array of size numpoints */
			double mmin;
			double mmax;
			double *ms;					/* array of size numpoints */
		} polygonz __attribute__ ((packed));

		struct {
			double box[4];
			int32_t numparts;
			int32_t numpoints;
			int32_t *parts;					/* array of size numparts */
			int32_t *parttypes;				/* array of size numparts */
			struct shapefile_point_struct *points;		/* array of size numpoints */
			double zmin;
			double zmax;
			double *zs;					/* array of size numpoints */
			double mmin;
			double mmax;
			double *ms;					/* array of size numpoints */
		} multipatch __attribute__ ((packed));

	} __attribute__ ((packed));
} __attribute__ ((packed));

class ShapefilePolyLinePart {
public:
	ShapefilePolyLinePart();
	~ShapefilePolyLinePart();

	std::list<struct shapefile_point_struct*> *GetPoints();
	bool ContainsPoint(double Lat, double Lon);
	void Clear();
private:
	std::list<struct shapefile_point_struct*> Points;
};

class ShapefilePolyLine {
public:
	ShapefilePolyLine();
	~ShapefilePolyLine();

	std::vector<ShapefilePolyLinePart*> *GetParts();
	void Clear();
	void ResetBoundingBox();
	void SetBoundingBox(double LatMin, double LonMin, double LatMax, double LonMax);
	void GetBoundingBox(double *LatMin, double *LonMin, double *LatMax, double *LonMax);
	bool InBoundingBox(double Lat, double Lon);	/* Is a point in this shape's bounding box? */
	/* Is our bounding box partially or completely within the given one? */
	bool ContainedIn(double LatMin, double LonMin, double LatMax, double LonMax);
	bool ContainsPoint(double Lat, double Lon);
	int32_t GetnPoints();

private:
	std::vector<ShapefilePolyLinePart*> Parts;
	double BoundingBox[4];	/* LatMin, LonMin, LatMax, LonMax */
};

/* A Polygon is virtually identical to a PolyLine */
class ShapefilePolygon: public ShapefilePolyLine {
public:
private:
};

class Shapefile {
public:
	Shapefile();
	~Shapefile();

	std::vector<ShapefilePolygon*> *GetPolygons();
	std::vector<ShapefilePolyLine*> *GetPolyLines();
	void GetPolyLineBounds(double *LatMin,double  *LonMin,double  *LatMax,double  *LonMax);
	void Clear();
	void Load(std::string Filename);
	void LoadPolygons(FILE *fp);
	void LoadPolyLines(FILE *fp);
	void SavePolyLines(std::string Filename);

private:
	std::vector<ShapefilePolygon*> Polygons;
	std::vector<ShapefilePolyLine*> PolyLines;
};

#endif	/* SHAPEFILE_H_ */

